
# This file was automatically generated by Dist::Zilla::Plugin::ModuleBuild v6.024.
use strict;
use warnings;

use Module::Build 0.28;


my %module_build_args = (
  "build_requires" => {
    "Module::Build" => "0.28"
  },
  "configure_requires" => {
    "Module::Build" => "0.28"
  },
  "dist_abstract" => "A Perl Solid Web Client",
  "dist_author" => [
    "Patrick Hochstenbach"
  ],
  "dist_name" => "Web-Solid-Auth",
  "dist_version" => "0.7",
  "license" => "perl",
  "module_name" => "Web::Solid::Auth",
  "recursive_test_files" => 1,
  "requires" => {
    "Attean" => 0,
    "Crypt::JWT" => 0,
    "Crypt::PK::ECC" => 0,
    "Data::UUID" => 0,
    "Digest::SHA" => 0,
    "File::LibMagic" => 0,
    "HTTP::Link" => 0,
    "JSON" => 0,
    "LWP::Protocol::https" => 0,
    "LWP::UserAgent" => 0,
    "Log::Any" => 0,
    "Log::Any::Adapter::Log4perl" => 0,
    "Log::Log4perl" => 0,
    "Moo" => 0,
    "Path::Tiny" => 0,
    "Plack" => 0,
    "String::Escape" => 0,
    "perl" => "v5.10.1"
  },
  "script_files" => [
    "bin/solid_auth.pl"
  ],
  "test_requires" => {
    "Dist::Zilla::Plugin::ExecDir" => 0,
    "Test::Exception" => "0.32",
    "Test::More" => "1.001003",
    "Test::Pod" => 0,
    "Test::Simple" => "1.001003"
  }
);


my %fallback_build_requires = (
  "Dist::Zilla::Plugin::ExecDir" => 0,
  "Module::Build" => "0.28",
  "Test::Exception" => "0.32",
  "Test::More" => "1.001003",
  "Test::Pod" => 0,
  "Test::Simple" => "1.001003"
);


unless ( eval { Module::Build->VERSION(0.4004) } ) {
  delete $module_build_args{test_requires};
  $module_build_args{build_requires} = \%fallback_build_requires;
}

my $build = Module::Build->new(%module_build_args);


$build->create_build_script;
